@@ -8,13 +8,13 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2010 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2011 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: DebuggerParser.cxx 2170 2010-11-07 18:10:58Z stephena $
+// $Id: DebuggerParser.cxx 2292 2011-12-22 21:29:24Z stephena $
 //============================================================================
 
 #include <fstream>
@@ -47,6 +47,8 @@
 #define CALL_METHOD(method) ( (this->*method)() )
 
 
+// TODO - use C++ streams instead of nasty C-strings and pointers
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 DebuggerParser::DebuggerParser(Debugger* d)
   : debugger(d)
@@ -137,12 +139,13 @@ string DebuggerParser::exec(const FilesystemNode& file)
 
     ostringstream buf;
     int count = 0;
-    char buffer[256];
+    string command;
     while( !in.eof() )
     {
-      if(!in.getline(buffer, 255))
+      if(!getline(in, command))
         break;
 
+      run(command);
       count++;
     }
     buf << "Executed " << debugger->valueToString(count) << " commands from \""
@@ -262,7 +265,11 @@ int DebuggerParser::decipher_arg(const string &str)
       } else { // must be hex.
         result = 0;
         while(*a != '\0') {
-          int hex = Debugger::conv_hex_digit(*a++);
+          int hex = -1;
+          char d = *a++;
+          if(d >= '0' && d <= '9')  hex = d - '0';
+          else if(d >= 'a' && d <= 'f') hex = d - 'a' + 10;
+          else if(d >= 'A' && d <= 'F') hex = d - 'A' + 10;
           if(hex < 0)
             return -1;
 
@@ -285,31 +292,25 @@ int DebuggerParser::decipher_arg(const string &str)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 string DebuggerParser::showWatches()
 {
-  string ret;
-  char buf[10];
-
-  for(unsigned int i=0; i<watches.size(); i++) {
-    if(watches[i] != "") {
+  ostringstream buf;
+  for(unsigned int i = 0; i < watches.size(); i++)
+  {
+    if(watches[i] != "")
+    {
       // Clear the args, since we're going to pass them to eval()
       argStrings.clear();
       args.clear();
 
-      sprintf(buf, "%d", i+1);
       argCount = 1;
       argStrings.push_back(watches[i]);
       args.push_back(decipher_arg(argStrings[0]));
-      if(args[0] < 0) {
-        ret += "BAD WATCH ";
-        ret += buf;
-        ret += ": " + argStrings[0] + "\n";
-      } else {
-        ret += " watch #";
-        ret += buf;
-        ret += " (" + argStrings[0] + ") -> " + eval() + "\n";
-      }
+      if(args[0] < 0)
+        buf << "BAD WATCH " << (i+1) << ": " << argStrings[0] << endl;
+      else
+        buf << " watch #" << (i+1) << " (" << argStrings[0] << ") -> " << eval() << endl;
     }
   }
-  return ret;
+  return buf.str();
 }
 
 
@@ -431,7 +432,7 @@ bool DebuggerParser::validateArgs(int cmd)
   while(*p != kARG_END_ARGS && *p != kARG_MULTI_BYTE)
   {
     count++;
-    *p++;
+    p++;
   }
 
   // Evil hack: some commands intentionally take multiple arguments
@@ -495,7 +496,7 @@ bool DebuggerParser::validateArgs(int cmd)
         break;
     }
     curCount++;
-    *p++;
+    p++;
 
   } while(*p != kARG_END_ARGS && curCount < argRequiredCount);
 
@@ -522,30 +523,38 @@ cerr << "curCount         = " << curCount << endl
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 string DebuggerParser::eval()
 {
-  char buf[50];
-  string ret;
-  for(int i=0; i<argCount; i++) {
-    // TODO - technically, we should determine if the label is read or write
-    const string& label = debugger->cartDebug().getLabel(args[i], true);
-    if(label != "") {
-      ret += label;
-      ret += ": ";
+  ostringstream buf;
+  for(int i = 0; i < argCount; ++i)
+  {
+    string rlabel = debugger->cartDebug().getLabel(args[i], true);
+    string wlabel = debugger->cartDebug().getLabel(args[i], false);
+    bool validR = rlabel != "" && rlabel[0] != '$',
+         validW = wlabel != "" && wlabel[0] != '$';
+    if(validR && validW)
+    {
+      if(rlabel == wlabel)
+        buf << rlabel << "(R/W): ";
+      else
+        buf << rlabel << "(R) / " << wlabel << "(W): ";
     }
-    ret += "$";
-    if(args[i] < 0x100) {
-      ret += Debugger::to_hex_8(args[i]);
-      ret += " %";
-      ret += Debugger::to_bin_8(args[i]);
-    } else {
-      ret += Debugger::to_hex_16(args[i]);
-      ret += " %";
-      ret += Debugger::to_bin_16(args[i]);
-    }
-    sprintf(buf, " #%d", args[i]);
-    ret += buf;
-    if(i != argCount - 1) ret += "\n";
+    else if(validR)
+      buf << rlabel << "(R): ";
+    else if(validW)
+      buf << wlabel << "(W): ";
+
+    if(args[i] < 0x100)
+      buf << "$" << debugger->valueToString(args[i], kBASE_16_2)
+          << " %" << debugger->valueToString(args[i], kBASE_2_8);
+    else
+      buf << "$" << debugger->valueToString(args[i], kBASE_16_4)
+          << " %" << debugger->valueToString(args[i], kBASE_2_16);
+
+    buf << " #" << (int) args[i]; 
+    if(i != argCount - 1)
+      buf << endl;
   }
-  return ret;
+
+  return buf.str();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -699,7 +708,7 @@ void DebuggerParser::executeBreak()
   else
     bp = args[0];
   debugger->toggleBreakPoint(bp);
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 
   if(debugger->breakPoint(bp))
     commandResult << "Set";
@@ -799,7 +808,7 @@ void DebuggerParser::executeClearwatches()
 // "cls"
 void DebuggerParser::executeCls()
 {
-  debugger->prompt()->clearScreen();
+  debugger->prompt().clearScreen();
   commandResult << "";
 }
 
@@ -822,7 +831,7 @@ void DebuggerParser::executeCode()
                   CartDebug::CODE, args[0], args[1]);
   commandResult << (result ? "added" : "removed") << " CODE directive on range $"
                 << hex << args[0] << " $" << hex << args[1];
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -863,7 +872,7 @@ void DebuggerParser::executeData()
                   CartDebug::DATA, args[0], args[1]);
   commandResult << (result ? "added" : "removed") << " DATA directive on range $"
                 << hex << args[0] << " $" << hex << args[1];
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -872,7 +881,7 @@ void DebuggerParser::executeDefine()
 {
   // TODO: check if label already defined?
   debugger->cartDebug().addLabel(argStrings[0], args[1]);
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -952,6 +961,13 @@ void DebuggerParser::executeExec()
   commandResult << exec(file);
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+// "exitrom"
+void DebuggerParser::executeExitRom()
+{
+  debugger->quit(true);
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 // "frame"
 void DebuggerParser::executeFrame()
@@ -1002,7 +1018,7 @@ void DebuggerParser::executeGfx()
                   CartDebug::GFX, args[0], args[1]);
   commandResult << (result ? "added" : "removed") << " GFX directive on range $"
                 << hex << args[0] << " $" << hex << args[1];
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1040,7 +1056,7 @@ void DebuggerParser::executeJump()
 
   if(line >= 0 && address >= 0)
   {
-    debugger->myRom->scrollTo(line);
+    debugger->rom().scrollTo(line);
     commandResult << "disassembly scrolled to address $" << HEX4 << address;
   }
   else
@@ -1141,7 +1157,7 @@ void DebuggerParser::executeLoadconfig()
   else
     commandResult << debugger->cartDebug().loadConfigFile();
 
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1159,7 +1175,7 @@ void DebuggerParser::executeLoadstate()
 void DebuggerParser::executeLoadsym()
 {
   commandResult << debugger->cartDebug().loadSymbolFile(argStrings[0]);
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1198,7 +1214,7 @@ void DebuggerParser::executePGfx()
                   CartDebug::PGFX, args[0], args[1]);
   commandResult << (result ? "added" : "removed") << " PGFX directive on range $"
                 << hex << args[0] << " $" << hex << args[1];
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1223,7 +1239,7 @@ void DebuggerParser::executeRam()
 void DebuggerParser::executeReset()
 {
   debugger->reset();
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
   commandResult << "reset CPU";
 }
 
@@ -1233,7 +1249,7 @@ void DebuggerParser::executeRewind()
 {
   if(debugger->rewindState())
   {
-    debugger->myRom->invalidate();
+    debugger->rom().invalidate();
     commandResult << "rewind by one level";
   }
   else
@@ -1266,7 +1282,7 @@ void DebuggerParser::executeRom()
   // The RomWidget is a special case, since we don't want to re-disassemble
   // any more than necessary.  So we only do it by calling the following
   // method ...
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 
   commandResult << "changed " << debugger->valueToString( args.size() - 1 )
                 << " location(s)";
@@ -1291,7 +1307,7 @@ void DebuggerParser::executeRow()
                   CartDebug::ROW, args[0], args[1]);
   commandResult << (result ? "added" : "removed") << " ROW directive on range $"
                 << hex << args[0] << " $" << hex << args[1];
-  debugger->myRom->invalidate();
+  debugger->rom().invalidate();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1299,7 +1315,7 @@ void DebuggerParser::executeRow()
 void DebuggerParser::executeRun()
 {
   debugger->saveOldState();
-  debugger->quit();
+  debugger->quit(false);
   commandResult << "_EXIT_DEBUGGER";  // See PromptWidget for more info
 }
 
@@ -1410,8 +1426,9 @@ void DebuggerParser::executeSaveconfig()
 // "saverom"
 void DebuggerParser::executeSaverom()
 {
-  if(debugger->saveROM(argStrings[0]))
-    commandResult << "saved ROM as " << argStrings[0];
+  const string& result = debugger->saveROM(argStrings[0]);
+  if(result != "")
+    commandResult << "saved ROM as " << result;
   else
     commandResult << red("failed to save ROM");
 }
@@ -1420,7 +1437,7 @@ void DebuggerParser::executeSaverom()
 // "saveses"
 void DebuggerParser::executeSaveses()
 {
-  if(debugger->prompt()->saveBuffer(argStrings[0]))
+  if(debugger->prompt().saveBuffer(argStrings[0]))
     commandResult << "saved session to file " << argStrings[0];
   else
     commandResult << red("I/O error");
@@ -1563,7 +1580,7 @@ void DebuggerParser::executeUndef()
 {
   if(debugger->cartDebug().removeLabel(argStrings[0]))
   {
-    debugger->myRom->invalidate();
+    debugger->rom().invalidate();
     commandResult << argStrings[0] + " now undefined";
   }
   else
@@ -1823,6 +1840,15 @@ DebuggerParser::Command DebuggerParser::commands[kNumCommands] = {
     &DebuggerParser::executeExec
   },
 
+  {
+    "exitrom",
+    "Exit emulator, return to ROM launcher",
+    false,
+    false,
+    { kARG_END_ARGS },
+    &DebuggerParser::executeExitRom
+  },
+
   {
     "frame",
     "Advance emulation by xx frames (default=1)",
