@@ -8,19 +8,23 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2010 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2011 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: CartDPCPlus.cxx 2139 2010-10-03 18:19:09Z stephena $
+// $Id: CartDPCPlus.cxx 2279 2011-11-07 22:50:23Z stephena $
 //============================================================================
 
 #include <cassert>
 #include <cstring>
 
+#ifdef DEBUGGER_SUPPORT
+  #include "Debugger.hxx"
+#endif
 #include "System.hxx"
+#include "Thumbulator.hxx"
 #include "CartDPCPlus.hxx"
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -29,7 +33,7 @@ CartridgeDPCPlus::CartridgeDPCPlus(const uInt8* image, uInt32 size,
   : Cartridge(settings),
     myFastFetch(false),
     myLDAimmediate(false),
-    myParameter(0),
+    myParameterPointer(0),
     mySystemCycles(0),
     myFractionalClocks(0.0)
 {
@@ -37,17 +41,19 @@ CartridgeDPCPlus::CartridgeDPCPlus(const uInt8* image, uInt32 size,
   uInt32 minsize = 4096 * 6 + 4096 + 1024 + 255;
   mySize = BSPF_max(minsize, size);
   myImage = new uInt8[mySize];
+  myDPCRAM = new uInt8[8192];
   memcpy(myImage, image, size);
   createCodeAccessBase(4096 * 6);
 
   // Pointer to the program ROM (24K @ 0 byte offset)
   myProgramImage = myImage;
 
-  // Pointer to the display ROM (4K @ 24K offset)
-  myDisplayImage = myProgramImage + 4096 * 6;
+  // Pointer to the display RAM
+  myDisplayImage = myDPCRAM + 0xC00;
+  memset(myDPCRAM, 0, 8192);
 
   // Pointer to the Frequency ROM (1K @ 28K offset)
-  myFrequencyImage = myDisplayImage + 4096;
+  myFrequencyImage = myProgramImage + 0x7000;
 
   // If the image is larger than 29K, we assume any excess at the
   // beginning is ARM code, and skip over it
@@ -55,13 +61,24 @@ CartridgeDPCPlus::CartridgeDPCPlus(const uInt8* image, uInt32 size,
   {
     int offset = size - 29 * 1024;
     myProgramImage   += offset;
-    myDisplayImage   += offset;
+//    myDisplayImage   += offset;
     myFrequencyImage += offset;
   }
 
+#ifdef THUMB_SUPPORT
+  // Create Thumbulator ARM emulator
+  myThumbEmulator = new Thumbulator((uInt16*)(myProgramImage-0xC00),
+                                    (uInt16*)myDPCRAM,
+                                     settings.getBool("thumb.trapfatal"));
+#endif
+
+  // Copy DPC display data to Harmony RAM
+  memcpy(myDisplayImage, myProgramImage + 0x6000, 0x1000);
+
   // Initialize the DPC data fetcher registers
   for(uInt16 i = 0; i < 8; ++i)
-    myTops[i] = myBottoms[i] = myCounters[i] = myFractionalIncrements[i] = 0;
+    myTops[i] = myBottoms[i] = myCounters[i] = myFractionalIncrements[i] = 
+    myFractionalCounters[i] = 0;
 
   // Set waveforms to first waveform entry
   myMusicWaveforms[0] = myMusicWaveforms[1] = myMusicWaveforms[2] = 0;
@@ -77,6 +94,11 @@ CartridgeDPCPlus::CartridgeDPCPlus(const uInt8* image, uInt32 size,
 CartridgeDPCPlus::~CartridgeDPCPlus()
 {
   delete[] myImage;
+  delete[] myDPCRAM;
+
+#ifdef THUMB_SUPPORT
+  delete myThumbEmulator;
+#endif
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -124,14 +146,17 @@ void CartridgeDPCPlus::install(System& system)
 inline void CartridgeDPCPlus::clockRandomNumberGenerator()
 {
   // Update random number generator (32-bit LFSR)
-  myRandomNumber = ((myRandomNumber & 1) ? 0xa260012b: 0x00) ^ ((myRandomNumber >> 1) & 0x7FFFFFFF);
+  myRandomNumber = ((myRandomNumber & (1<<10)) ? 0x10adab1e: 0x00) ^
+                   ((myRandomNumber >> 11) | (myRandomNumber << 21));
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 inline void CartridgeDPCPlus::priorClockRandomNumberGenerator()
 {
   // Update random number generator (32-bit LFSR, reversed)
-  myRandomNumber = ((myRandomNumber & (1<<31)) ? 0x44c00257: 0x00) ^ (myRandomNumber << 1);
+  myRandomNumber = ((myRandomNumber & (1<<31)) ?
+    ((0x10adab1e^myRandomNumber) << 11) | ((0x10adab1e^myRandomNumber) >> 21) :
+    (myRandomNumber << 11) | (myRandomNumber >> 21));
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -161,15 +186,42 @@ inline void CartridgeDPCPlus::updateMusicModeDataFetchers()
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 inline void CartridgeDPCPlus::callFunction(uInt8 value)
 {
-  //  myParameter
+  // myParameter
+  uInt16 ROMdata = (myParameter[1] << 8) + myParameter[0];
   switch (value)
   {
-    //case 0:     template for future special functions
-      // CallSpecialFunctionO(myParameter);
-      // break;
-    case 255: // Call user writen ARM code (most likely be C compiled for ARM).
-              // ARM code not supported by Stella at this time.
+    case 0: // Parameter Pointer reset
+      myParameterPointer = 0;
       break;
+    case 1: // Copy ROM to fetcher
+      for(int i = 0; i < myParameter[3]; ++i)
+        myDisplayImage[myCounters[myParameter[2] & 0x7]+i] = myProgramImage[ROMdata+i];
+      myParameterPointer = 0;
+      break;
+    case 2: // Copy value to fetcher
+      for(int i = 0; i < myParameter[3]; ++i)
+        myDisplayImage[myCounters[myParameter[2]]+i] = myParameter[0];
+      myParameterPointer = 0;
+      break;
+  #ifdef THUMB_SUPPORT
+    case 254:
+    case 255:
+      // Call user written ARM code (most likely be C compiled for ARM)
+      try {
+        myThumbEmulator->run();
+      }
+      catch(const string& error) {
+        if(!mySystem->autodectMode())
+        {
+      #ifdef DEBUGGER_SUPPORT
+          Debugger::debugger().startWithFatalError(error);
+      #else
+          cout << error << endl;
+      #endif
+        }
+      }
+      break;
+  #endif
     // reserved
   }
 }
@@ -240,11 +292,11 @@ uInt8 CartridgeDPCPlus::peek(uInt16 address)
             // Update the music data fetchers (counter & flag)
             updateMusicModeDataFetchers();
 
-            // using myProgramImage[] instead of myDisplayImage[] because waveforms
-            // could also be in the 1K Frequency table.
-            uInt32 i = myProgramImage[6*4096 + (myMusicWaveforms[0] << 5) + (myMusicCounters[0] >> 27)] +
-                       myProgramImage[6*4096 + (myMusicWaveforms[1] << 5) + (myMusicCounters[1] >> 27)] +
-                       myProgramImage[6*4096 + (myMusicWaveforms[2] << 5) + (myMusicCounters[2] >> 27)];
+            // using myDisplayImage[] instead of myProgramImage[] because waveforms
+            // can be modified during runtime.
+            uInt32 i = myDisplayImage[(myMusicWaveforms[0] << 5) + (myMusicCounters[0] >> 27)] +
+                       myDisplayImage[(myMusicWaveforms[1] << 5) + (myMusicCounters[1] >> 27)] +
+                       myDisplayImage[(myMusicWaveforms[2] << 5) + (myMusicCounters[2] >> 27)];
 
             result = (uInt8)i;
             break;
@@ -409,7 +461,8 @@ bool CartridgeDPCPlus::poke(uInt16 address, uInt8 value)
             break;
 
           case 0x01:  // PARAMETER - set parameter used by CALLFUNCTION (not all functions use the parameter)
-            myParameter = value;
+            if(myParameterPointer < 8)
+              myParameter[myParameterPointer++] = value;
             break;
 
           case 0x02:  // CALLFUNCTION
@@ -642,7 +695,9 @@ bool CartridgeDPCPlus::save(Serializer& out) const
     out.putBool(myLDAimmediate);
 
     // Control Byte to update
-    out.putByte((char) myParameter);
+    out.putInt(8);
+    for(i = 0; i < 8; ++i)
+      out.putByte((char)myParameter[i]);
 
     // The music counters
     out.putInt(3);
@@ -717,7 +772,9 @@ bool CartridgeDPCPlus::load(Serializer& in)
     myLDAimmediate = in.getBool();
 
     // Control Byte to update
-    myParameter = (uInt8) in.getByte();
+    limit = (uInt32) in.getInt();
+    for(i = 0; i < limit; ++i)
+      myParameter[i] = (uInt8) in.getByte();
 
     // The music mode counters for the data fetchers
     limit = (uInt32) in.getInt();
