@@ -8,12 +8,13 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2009 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2011 by Bradford W. Mott, Stephen Anthony
+// and the Stella Team
 //
-// See the file "license" for information on usage and redistribution of
+// See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: OSystem.cxx 1824 2009-07-01 16:04:28Z stephena $
+// $Id: OSystem.cxx 2312 2011-12-29 19:10:57Z stephena $
 //============================================================================
 
 #include <cassert>
@@ -21,14 +22,15 @@
 #include <fstream>
 #include <zlib.h>
 
+#include <ctime>
 #ifdef HAVE_GETTIMEOFDAY
-  #include <time.h>
   #include <sys/time.h>
 #endif
 
 #include "bspf.hxx"
 
 #include "MediaFactory.hxx"
+#include "Sound.hxx"
 
 #ifdef DEBUGGER_SUPPORT
   #include "Debugger.hxx"
@@ -50,6 +52,7 @@
 #include "FSNode.hxx"
 #include "unzip.h"
 #include "MD5.hxx"
+#include "Cart.hxx"
 #include "Settings.hxx"
 #include "PropsSet.hxx"
 #include "EventHandler.hxx"
@@ -65,6 +68,7 @@
 #include "Console.hxx"
 #include "Random.hxx"
 #include "StateManager.hxx"
+#include "Version.hxx"
 
 #include "OSystem.hxx"
 
@@ -89,54 +93,70 @@ OSystem::OSystem()
     myRomFile(""),
     myRomMD5(""),
     myFeatures(""),
+    myBuildInfo(""),
     myFont(NULL),
     myConsoleFont(NULL)
 {
-#ifdef DISPLAY_OPENGL
-  myFeatures += "OpenGL ";
-#endif
-#ifdef SOUND_SUPPORT
-  myFeatures += "Sound ";
-#endif
-#ifdef JOYSTICK_SUPPORT
-  myFeatures += "Joystick ";
-#endif
-#ifdef DEBUGGER_SUPPORT
-  myFeatures += "Debugger ";
-#endif
-#ifdef CHEATCODE_SUPPORT
-  myFeatures += "Cheats";
-#endif
+  // Calculate startup time
+  myMillisAtStart = (uInt32)(time(NULL) * 1000);
+
+  // Get built-in features
+  #ifdef DISPLAY_OPENGL
+    myFeatures += "OpenGL ";
+  #endif
+  #ifdef SOUND_SUPPORT
+    myFeatures += "Sound ";
+  #endif
+  #ifdef JOYSTICK_SUPPORT
+    myFeatures += "Joystick ";
+  #endif
+  #ifdef DEBUGGER_SUPPORT
+    myFeatures += "Debugger ";
+  #endif
+  #ifdef CHEATCODE_SUPPORT
+    myFeatures += "Cheats";
+  #endif
+
+  // Get build info
+  ostringstream info;
+  const SDL_version* ver = SDL_Linked_Version();
+
+  info << "Build " << STELLA_BUILD << ", using ";
+  info << "SDL " << (int)ver->major << "." << (int)ver->minor << "." << (int)ver->patch << " ";
+  info << "[" << BSPF_ARCH << "]";
+  myBuildInfo = info.str();
 
 #if 0
   // Debugging info for the GUI widgets
-  cerr << "  kStaticTextWidget   = " << kStaticTextWidget   << endl;
-  cerr << "  kEditTextWidget     = " << kEditTextWidget     << endl;
-  cerr << "  kButtonWidget       = " << kButtonWidget       << endl;
-  cerr << "  kCheckboxWidget     = " << kCheckboxWidget     << endl;
-  cerr << "  kSliderWidget       = " << kSliderWidget       << endl;
-  cerr << "  kListWidget         = " << kListWidget         << endl;
-  cerr << "  kScrollBarWidget    = " << kScrollBarWidget    << endl;
-  cerr << "  kPopUpWidget        = " << kPopUpWidget        << endl;
-  cerr << "  kTabWidget          = " << kTabWidget	        << endl;
-  cerr << "  kEventMappingWidget = " << kEventMappingWidget << endl;
-  cerr << "  kEditableWidget     = " << kEditableWidget     << endl;
-  cerr << "  kAudioWidget        = " << kAudioWidget        << endl;
-  cerr << "  kColorWidget        = " << kColorWidget        << endl;
-  cerr << "  kCpuWidget          = " << kCpuWidget          << endl;
-  cerr << "  kDataGridOpsWidget  = " << kDataGridOpsWidget  << endl;
-  cerr << "  kDataGridWidget     = " << kDataGridWidget     << endl;
-  cerr << "  kPromptWidget       = " << kPromptWidget       << endl;
-  cerr << "  kRamWidget          = " << kRamWidget          << endl;
-  cerr << "  kRomListWidget      = " << kRomListWidget      << endl;
-  cerr << "  kRomWidget          = " << kRomWidget          << endl;
-  cerr << "  kTiaInfoWidget      = " << kTiaInfoWidget      << endl;
-  cerr << "  kTiaOutputWidget    = " << kTiaOutputWidget    << endl;
-  cerr << "  kTiaWidget          = " << kTiaWidget          << endl;
-  cerr << "  kTiaZoomWidget      = " << kTiaZoomWidget      << endl;
-  cerr << "  kToggleBitWidget    = " << kToggleBitWidget    << endl;
-  cerr << "  kTogglePixelWidget  = " << kTogglePixelWidget  << endl;
-  cerr << "  kToggleWidget       = " << kToggleWidget       << endl;
+  ostringstream buf;
+  buf << "  kStaticTextWidget   = " << kStaticTextWidget   << endl
+      << "  kEditTextWidget     = " << kEditTextWidget     << endl
+      << "  kButtonWidget       = " << kButtonWidget       << endl
+      << "  kCheckboxWidget     = " << kCheckboxWidget     << endl
+      << "  kSliderWidget       = " << kSliderWidget       << endl
+      << "  kListWidget         = " << kListWidget         << endl
+      << "  kScrollBarWidget    = " << kScrollBarWidget    << endl
+      << "  kPopUpWidget        = " << kPopUpWidget        << endl
+      << "  kTabWidget          = " << kTabWidget          << endl
+      << "  kEventMappingWidget = " << kEventMappingWidget << endl
+      << "  kEditableWidget     = " << kEditableWidget     << endl
+      << "  kAudioWidget        = " << kAudioWidget        << endl
+      << "  kColorWidget        = " << kColorWidget        << endl
+      << "  kCpuWidget          = " << kCpuWidget          << endl
+      << "  kDataGridOpsWidget  = " << kDataGridOpsWidget  << endl
+      << "  kDataGridWidget     = " << kDataGridWidget     << endl
+      << "  kPromptWidget       = " << kPromptWidget       << endl
+      << "  kRamWidget          = " << kRamWidget          << endl
+      << "  kRomListWidget      = " << kRomListWidget      << endl
+      << "  kRomWidget          = " << kRomWidget          << endl
+      << "  kTiaInfoWidget      = " << kTiaInfoWidget      << endl
+      << "  kTiaOutputWidget    = " << kTiaOutputWidget    << endl
+      << "  kTiaWidget          = " << kTiaWidget          << endl
+      << "  kTiaZoomWidget      = " << kTiaZoomWidget      << endl
+      << "  kToggleBitWidget    = " << kToggleBitWidget    << endl
+      << "  kTogglePixelWidget  = " << kTogglePixelWidget  << endl
+      << "  kToggleWidget       = " << kToggleWidget       << endl;
+  logMessage(buf.str(), 0);
 #endif
 }
 
@@ -147,6 +167,7 @@ OSystem::~OSystem()
   delete myCommandMenu;
   delete myLauncher;
   delete myFont;
+  delete myInfoFont;
   delete mySmallFont;
   delete myConsoleFont;
   delete myLauncherFont;
@@ -181,6 +202,18 @@ bool OSystem::create()
 {
   // Get updated paths for all configuration files
   setConfigPaths();
+  ostringstream buf;
+  buf << "Stella " << STELLA_VERSION << endl
+      << "  Features: " << myFeatures << endl
+      << "  " << myBuildInfo << endl << endl
+      << "Base directory:       '"
+      << FilesystemNode(myBaseDir).getRelativePath() << "'" << endl
+      << "Configuration file:   '"
+      << FilesystemNode(myConfigFile).getRelativePath() << "'" << endl
+      << "User game properties: '"
+      << FilesystemNode(myPropertiesFile).getRelativePath() << "'" << endl
+      << endl;
+  logMessage(buf.str(), 1);
 
   // Get relevant information about the video hardware
   // This must be done before any graphics context is created, since
@@ -200,18 +233,22 @@ bool OSystem::create()
   ////////////////////////////////////////////////////////////////////
   bool smallScreen = myDesktopWidth < 640 || myDesktopHeight < 480;
 
+  // The console font is always the same size (for now at least)
+  myConsoleFont  = new GUI::Font(GUI::consoleDesc);
+
   // This font is used in a variety of situations when a really small
   // font is needed; we let the specific widget/dialog decide when to
   // use it
   mySmallFont = new GUI::Font(GUI::stellaDesc);
 
-  // The console font is always the same size (for now at least)
-  myConsoleFont  = new GUI::Font(GUI::consoleDesc);
-
   // The general font used in all UI elements
   // This is determined by the size of the framebuffer
   myFont = new GUI::Font(smallScreen ? GUI::stellaDesc : GUI::stellaMediumDesc);
 
+  // The info font used in all UI elements
+  // This is determined by the size of the framebuffer
+  myInfoFont = new GUI::Font(smallScreen ? GUI::stellaDesc : GUI::consoleDesc);
+
   // The font used by the ROM launcher
   // Normally, this is configurable by the user, except in the case of
   // very small screens
@@ -282,44 +319,25 @@ void OSystem::setConfigPaths()
   FilesystemNode node;
   string s;
 
-  s = mySettings->getString("statedir");
-  if(s == "") s = myBaseDir + BSPF_PATH_SEPARATOR + "state";
-  node = FilesystemNode(s);
-  myStateDir = node.getPath();
-  mySettings->setString("statedir", node.getRelativePath());
-  if(!node.isDirectory())
-    AbstractFilesystemNode::makeDir(myStateDir);
-
-  s = mySettings->getString("ssdir");
-  if(s == "") s = myBaseDir + BSPF_PATH_SEPARATOR + "snapshots";
-  node = FilesystemNode(s);
-  mySnapshotDir = node.getPath();
-  mySettings->setString("ssdir", node.getRelativePath());
-  if(!node.isDirectory())
-    AbstractFilesystemNode::makeDir(mySnapshotDir);
-
-  s = mySettings->getString("eepromdir");
-  if(s == "") s = myBaseDir;
-  node = FilesystemNode(s);
-  myEEPROMDir = node.getPath();
-  mySettings->setString("eepromdir", node.getRelativePath());
-  if(!node.isDirectory())
-    AbstractFilesystemNode::makeDir(myEEPROMDir);
+  validatePath(myStateDir, "statedir", myBaseDir + "statedir");
+  validatePath(mySnapshotDir, "snapdir", defaultSnapDir());
+  validatePath(myEEPROMDir, "eepromdir", myBaseDir);
+  validatePath(myCfgDir, "cfgdir", myBaseDir + "cfg");
 
   s = mySettings->getString("cheatfile");
-  if(s == "") s = myBaseDir + BSPF_PATH_SEPARATOR + "stella.cht";
+  if(s == "") s = myBaseDir + "stella.cht";
   node = FilesystemNode(s);
   myCheatFile = node.getPath();
   mySettings->setString("cheatfile", node.getRelativePath());
 
   s = mySettings->getString("palettefile");
-  if(s == "") s = myBaseDir + BSPF_PATH_SEPARATOR + "stella.pal";
+  if(s == "") s = myBaseDir + "stella.pal";
   node = FilesystemNode(s);
   myPaletteFile = node.getPath();
   mySettings->setString("palettefile", node.getRelativePath());
 
   s = mySettings->getString("propsfile");
-  if(s == "") s = myBaseDir + BSPF_PATH_SEPARATOR + "stella.pro";
+  if(s == "") s = myBaseDir + "stella.pro";
   node = FilesystemNode(s);
   myPropertiesFile = node.getPath();
   mySettings->setString("propsfile", node.getRelativePath());
@@ -340,7 +358,10 @@ void OSystem::setBaseDir(const string& basedir)
   FilesystemNode node(basedir);
   myBaseDir = node.getPath();
   if(!node.isDirectory())
+  {
     AbstractFilesystemNode::makeDir(myBaseDir);
+    myBaseDir = FilesystemNode(node.getPath()).getPath();
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -353,12 +374,15 @@ void OSystem::setConfigFile(const string& file)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void OSystem::setFramerate(float framerate)
 {
-  myDisplayFrameRate = framerate;
-  myTimePerFrame = (uInt32)(1000000.0 / myDisplayFrameRate);
+  if(framerate > 0.0)
+  {
+    myDisplayFrameRate = framerate;
+    myTimePerFrame = (uInt32)(1000000.0 / myDisplayFrameRate);
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool OSystem::createFrameBuffer()
+FBInitStatus OSystem::createFrameBuffer()
 {
   // There is only ever one FrameBuffer created per run of Stella
   // Due to the multi-surface nature of the FrameBuffer, repeatedly
@@ -373,30 +397,34 @@ bool OSystem::createFrameBuffer()
     myFrameBuffer = MediaFactory::createVideo(this);
 
   // Re-initialize the framebuffer to current settings
+  FBInitStatus fbstatus = kFailComplete;
   switch(myEventHandler->state())
   {
     case EventHandler::S_EMULATE:
     case EventHandler::S_PAUSE:
     case EventHandler::S_MENU:
     case EventHandler::S_CMDMENU:
-      if(!myConsole->initializeVideo())
+      fbstatus = myConsole->initializeVideo();
+      if(fbstatus != kSuccess)
         goto fallback;
       break;  // S_EMULATE, S_PAUSE, S_MENU, S_CMDMENU
 
     case EventHandler::S_LAUNCHER:
-      if(!myLauncher->initializeVideo())
+      fbstatus = myLauncher->initializeVideo();
+      if(fbstatus != kSuccess)
         goto fallback;
       break;  // S_LAUNCHER
 
 #ifdef DEBUGGER_SUPPORT
     case EventHandler::S_DEBUGGER:
-      if(!myDebugger->initializeVideo())
+      fbstatus = myDebugger->initializeVideo();
+      if(fbstatus != kSuccess)
         goto fallback;
       break;  // S_DEBUGGER
 #endif
 
     default:  // Should never happen
-      cerr << "ERROR: Unknown emulation state in createFrameBuffer()" << endl;
+      logMessage("ERROR: Unknown emulation state in createFrameBuffer()\n", 0);
       break;
   }
 
@@ -410,34 +438,37 @@ bool OSystem::createFrameBuffer()
     setUIPalette();
   }
 
-  return true;
+  return fbstatus;
 
   // GOTO are normally considered evil, unless well documented :)
-  // If initialization of video system fails while in OpenGL mode,
-  // attempt to fallback to software mode
+  // If initialization of video system fails while in OpenGL mode
+  // because OpenGL is unavailable, attempt to fallback to software mode
+  // Otherwise, pass the error to the parent
 fallback:
-  if(myFrameBuffer && myFrameBuffer->type() == kGLBuffer)
+  if(fbstatus == kFailNotSupported && myFrameBuffer &&
+     myFrameBuffer->type() == kDoubleBuffer)
   {
-    cerr << "ERROR: OpenGL mode failed, fallback to software" << endl;
+    logMessage("ERROR: OpenGL mode failed, fallback to software\n", 0);
     delete myFrameBuffer; myFrameBuffer = NULL;
     mySettings->setString("video", "soft");
-    bool ret = createFrameBuffer();
-    if(ret)
+    FBInitStatus newstatus = createFrameBuffer();
+    if(newstatus == kSuccess)
     {
       setFramerate(60);
-      myFrameBuffer->showMessage("OpenGL mode failed, fallback to software", kMiddleCenter);
+      myFrameBuffer->showMessage("OpenGL mode failed, fallback to software",
+                                 kMiddleCenter, true);
     }
-    return ret;
+    return newstatus;
   }
   else
-    return false;
+    return fbstatus;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void OSystem::createSound()
 {
-  delete mySound;  mySound = NULL;
-  mySound = MediaFactory::createAudio(this);
+  if(!mySound)
+    mySound = MediaFactory::createAudio(this);
 #ifndef SOUND_SUPPORT
   mySettings->setBool("sound", false);
 #endif
@@ -446,10 +477,12 @@ void OSystem::createSound()
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool OSystem::createConsole(const string& romfile, const string& md5sum)
 {
+  ostringstream buf;
+
   // Do a little error checking; it shouldn't be necessary
   if(myConsole) deleteConsole();
 
-  bool retval = false, showmessage = false;
+  bool showmessage = false;
 
   // If a blank ROM has been given, we reload the current one (assuming one exists)
   if(romfile == "")
@@ -457,7 +490,7 @@ bool OSystem::createConsole(const string& romfile, const string& md5sum)
     showmessage = true;  // we show a message if a ROM is being reloaded
     if(myRomFile == "")
     {
-      cerr << "ERROR: Rom file not specified ..." << endl;
+      logMessage("ERROR: Rom file not specified ...\n", 0);
       return false;
     }
   }
@@ -473,13 +506,13 @@ bool OSystem::createConsole(const string& romfile, const string& md5sum)
   }
 
   // Create an instance of the 2600 game console
-  myConsole = openConsole(myRomFile, myRomMD5);
+  string type, id;
+  myConsole = openConsole(myRomFile, myRomMD5, type, id);
   if(myConsole)
   {
   #ifdef CHEATCODE_SUPPORT
     myCheatManager->loadCheats(myRomMD5);
   #endif
-    bool audiofirst = mySettings->getBool("audiofirst");
     //////////////////////////////////////////////////////////////////////////
     // For some reason, ATI video drivers for OpenGL in Win32 cause problems
     // if the sound isn't initialized before the video
@@ -489,49 +522,50 @@ bool OSystem::createConsole(const string& romfile, const string& md5sum)
     // http://www.atariage.com/forums/index.php?showtopic=126090&view=findpost&p=1648693
     // Hopefully it won't break anything else
     //////////////////////////////////////////////////////////////////////////
-    if(audiofirst)  myConsole->initializeAudio();
+    myConsole->initializeAudio();
     myEventHandler->reset(EventHandler::S_EMULATE);
-    if(!createFrameBuffer())  // Takes care of initializeVideo()
+    myEventHandler->setMouseControllerMode(mySettings->getString("mcontrol"));
+    if(createFrameBuffer() != kSuccess)  // Takes care of initializeVideo()
     {
-      cerr << "ERROR: Couldn't create framebuffer for console" << endl;
+      logMessage("ERROR: Couldn't create framebuffer for console\n", 0);
       myEventHandler->reset(EventHandler::S_LAUNCHER);
       return false;
     }
-    if(!audiofirst)  myConsole->initializeAudio();
-  #ifdef DEBUGGER_SUPPORT
-    myDebugger->setConsole(myConsole);
-    myDebugger->initialize();
-  #endif
 
     if(showmessage)
-      myFrameBuffer->showMessage("New console created");
-    if(mySettings->getBool("showinfo"))
-      cout << "Game console created:" << endl
-           << "  ROM file: " << myRomFile << endl << endl
-           << getROMInfo(myConsole) << endl;
+    {
+      if(id == "")
+        myFrameBuffer->showMessage("New console created");
+      else
+        myFrameBuffer->showMessage("Multicart " + type + ", loading ROM" + id);
+    }
+    buf << "Game console created:" << endl
+        << "  ROM file: " << FilesystemNode(myRomFile).getRelativePath() << endl << endl
+        << getROMInfo(myConsole) << endl;
+    logMessage(buf.str(), 1);
 
     // Update the timing info for a new console run
     resetLoopTiming();
 
     myFrameBuffer->setCursorState();
-    retval = true;
+
+    // Also check if certain virtual buttons should be held down
+    // These must be checked each time a new console is being created
+    if(mySettings->getBool("holdreset"))
+      myEventHandler->handleEvent(Event::ConsoleReset, 1);
+    if(mySettings->getBool("holdselect"))
+      myEventHandler->handleEvent(Event::ConsoleSelect, 1);
+    if(mySettings->getBool("holdbutton0"))
+      myEventHandler->handleEvent(Event::JoystickZeroFire1, 1);
+
+    return true;
   }
   else
   {
-    cerr << "ERROR: Couldn't create console for " << myRomFile << endl;
-    retval = false;
+    buf << "ERROR: Couldn't create console for " << myRomFile << endl;
+    logMessage(buf.str(), 0);
+    return false;
   }
-
-  // Also check if certain virtual buttons should be held down
-  // These must be checked each time a new console is being created
-  if(mySettings->getBool("holdreset"))
-    myEventHandler->handleEvent(Event::ConsoleReset, 1);
-  if(mySettings->getBool("holdselect"))
-    myEventHandler->handleEvent(Event::ConsoleSelect, 1);
-  if(mySettings->getBool("holdbutton0"))
-    myEventHandler->handleEvent(Event::JoystickZeroFire1, 1);
-
-  return retval;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -543,16 +577,16 @@ void OSystem::deleteConsole()
   #ifdef CHEATCODE_SUPPORT
     myCheatManager->saveCheats(myConsole->properties().get(Cartridge_MD5));
   #endif
-    if(mySettings->getBool("showinfo"))
-    {
-      double executionTime   = (double) myTimingInfo.totalTime / 1000000.0;
-      double framesPerSecond = (double) myTimingInfo.totalFrames / executionTime;
-      cout << "Game console stats:" << endl
-           << "  Total frames drawn: " << myTimingInfo.totalFrames << endl
-           << "  Total time (sec):   " << executionTime << endl
-           << "  Frames per second:  " << framesPerSecond << endl
-           << endl;
-    }
+    ostringstream buf;
+    double executionTime   = (double) myTimingInfo.totalTime / 1000000.0;
+    double framesPerSecond = (double) myTimingInfo.totalFrames / executionTime;
+    buf << "Game console stats:" << endl
+        << "  Total frames drawn: " << myTimingInfo.totalFrames << endl
+        << "  Total time (sec):   " << executionTime << endl
+        << "  Frames per second:  " << framesPerSecond << endl
+        << endl;
+    logMessage(buf.str(), 1);
+
     delete myConsole;  myConsole = NULL;
   }
 }
@@ -561,9 +595,9 @@ void OSystem::deleteConsole()
 bool OSystem::createLauncher()
 {
   myEventHandler->reset(EventHandler::S_LAUNCHER);
-  if(!createFrameBuffer())
+  if(createFrameBuffer() != kSuccess)
   {
-    cerr << "ERROR: Couldn't create launcher" << endl;
+    logMessage("ERROR: Couldn't create launcher\n", 0);
     return false;
   }
   myLauncher->reStack();
@@ -579,8 +613,8 @@ bool OSystem::createLauncher()
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 string OSystem::getROMInfo(const string& romfile)
 {
-  string md5, result = "";
-  Console* console = openConsole(romfile, md5);
+  string md5, type, id, result = "";
+  Console* console = openConsole(romfile, md5, type, id);
   if(console)
   {
     result = getROMInfo(console);
@@ -607,7 +641,24 @@ string OSystem::MD5FromFile(const string& filename)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Console* OSystem::openConsole(const string& romfile, string& md5)
+void OSystem::logMessage(const string& message, uInt8 level)
+{
+  if(level == 0)
+  {
+    cout << message << flush;
+    myLogMessages += message;
+  }
+  else if(level <= (uInt8)mySettings->getInt("loglevel"))
+  {
+    if(mySettings->getBool("logtoconsole"))
+      cout << message << flush;
+    myLogMessages += message;
+  }
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Console* OSystem::openConsole(const string& romfile, string& md5,
+                              string& type, string& id)
 {
 #define CMDLINE_PROPS_UPDATE(cl_name, prop_name) \
   s = mySettings->getString(cl_name);            \
@@ -629,9 +680,10 @@ Console* OSystem::openConsole(const string& romfile, string& md5)
     CMDLINE_PROPS_UPDATE("type", Cartridge_Type);
 
     // Now create the cartridge
-    string id, cartmd5 = md5, type = props.get(Cartridge_Type);
+    string cartmd5 = md5;
+    type = props.get(Cartridge_Type);
     Cartridge* cart =
-      Cartridge::create(image, size, cartmd5, id, type, *mySettings);
+      Cartridge::create(image, size, cartmd5, type, id, *mySettings);
 
     // It's possible that the cart created was from a piece of the image,
     // and that the md5 (and hence the cart) has changed
@@ -668,7 +720,11 @@ Console* OSystem::openConsole(const string& romfile, string& md5)
       console = new Console(this, cart, props);
   }
   else
-    cerr << "ERROR: Couldn't open " << romfile << endl;
+  {
+    ostringstream buf;
+    buf << "ERROR: Couldn't open \'" << romfile << "\'" << endl;
+    logMessage(buf.str(), 0);
+  }
 
   // Free the image since we don't need it any longer
   if(image != 0 && size > 0)
@@ -708,10 +764,10 @@ uInt8* OSystem::openROM(string file, string& md5, uInt32& size)
         if(strlen(filename) >= 4)
         {
           // Grab 3-character extension
-          char* ext = filename + strlen(filename) - 4;
+          const char* ext = filename + strlen(filename) - 4;
 
-          if(!BSPF_strcasecmp(ext, ".a26") || !BSPF_strcasecmp(ext, ".bin") ||
-             !BSPF_strcasecmp(ext, ".rom"))
+          if(BSPF_equalsIgnoreCase(ext, ".a26") || BSPF_equalsIgnoreCase(ext, ".bin") ||
+             BSPF_equalsIgnoreCase(ext, ".rom"))
           {
             file = filename;
             break;
@@ -734,7 +790,7 @@ uInt8* OSystem::openROM(string file, string& md5, uInt32& size)
 
       // We don't have to check for any return errors from these functions,
       // since if there are, 'image' will not contain a valid ROM and the
-      // calling method can take of it
+      // calling method can take care of it
       unzOpenCurrentFile(tz);
       unzReadCurrentFile(tz, image, size);
       unzCloseCurrentFile(tz);
@@ -758,6 +814,13 @@ uInt8* OSystem::openROM(string file, string& md5, uInt32& size)
     gzclose(f);
   }
 
+  // Zero-byte files should be automatically discarded
+  if(size == 0)
+  {
+    delete[] image;
+    return (uInt8*) 0;
+  }
+
   // If we get to this point, we know we have a valid file to open
   // Now we make sure that the file has a valid properties entry
   // To save time, only generate an MD5 if we really need one
@@ -771,8 +834,8 @@ uInt8* OSystem::openROM(string file, string& md5, uInt32& size)
   if(!myPropSet->getMD5(md5, props))
   {
     // Get the filename from the rom pathname
-    string::size_type pos = file.find_last_of("/\\");
-    if(pos != string::npos)  file = file.substr(pos+1);
+    FilesystemNode node(file);
+    file = node.getDisplayName();
 
     props.set(Cartridge_MD5, md5);
     props.set(Cartridge_Name, file);
@@ -801,52 +864,95 @@ string OSystem::getROMInfo(const Console* console)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void OSystem::resetLoopTiming()
 {
-  memset(&myTimingInfo, 0, sizeof(TimingInfo));
   myTimingInfo.start = myTimingInfo.virt = getTicks();
+  myTimingInfo.current = 0;
+  myTimingInfo.totalTime = 0;
+  myTimingInfo.totalFrames = 0;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void OSystem::setDefaultJoymap()
+void OSystem::validatePath(string& path, const string& setting,
+                           const string& defaultpath)
 {
-  EventMode mode;
-
-  mode = kEmulationMode;  // Default emulation events
-  // Left joystick (assume joystick zero, button zero)
-  myEventHandler->setDefaultJoyMapping(Event::JoystickZeroFire1, mode, 0, 0);
-  // Right joystick (assume joystick one, button zero)
-  myEventHandler->setDefaultJoyMapping(Event::JoystickOneFire1, mode, 1, 0);
+  const string& s = mySettings->getString(setting) == "" ? defaultpath :
+                    mySettings->getString(setting);
+  FilesystemNode node(s);
+  if(!node.isDirectory())
+  {
+    AbstractFilesystemNode::makeDir(s);
+    node = FilesystemNode(node.getPath());
+  }
+  path = node.getPath();
+  mySettings->setString(setting, node.getRelativePath());
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void OSystem::setDefaultJoyAxisMap()
+void OSystem::setDefaultJoymap(Event::Type event, EventMode mode)
 {
-  EventMode mode;
+#define SET_DEFAULT_AXIS(sda_event, sda_mode, sda_stick, sda_axis, sda_val, sda_cmp_event) \
+  if(eraseAll || sda_cmp_event == sda_event) \
+    myEventHandler->addJoyAxisMapping(sda_event, sda_mode, sda_stick, sda_axis, sda_val, false);
 
-  mode = kEmulationMode;  // Default emulation events
-  // Left joystick left/right directions (assume joystick zero)
-  myEventHandler->setDefaultJoyAxisMapping(Event::JoystickZeroLeft, mode, 0, 0, 0);
-  myEventHandler->setDefaultJoyAxisMapping(Event::JoystickZeroRight, mode, 0, 0, 1);
-  // Left joystick up/down directions (assume joystick zero)
-  myEventHandler->setDefaultJoyAxisMapping(Event::JoystickZeroUp, mode, 0, 1, 0);
-  myEventHandler->setDefaultJoyAxisMapping(Event::JoystickZeroDown, mode, 0, 1, 1);
-  // Right joystick left/right directions (assume joystick one)
-  myEventHandler->setDefaultJoyAxisMapping(Event::JoystickOneLeft, mode, 1, 0, 0);
-  myEventHandler->setDefaultJoyAxisMapping(Event::JoystickOneRight, mode, 1, 0, 1);
-  // Right joystick left/right directions (assume joystick one)
-  myEventHandler->setDefaultJoyAxisMapping(Event::JoystickOneUp, mode, 1, 1, 0);
-  myEventHandler->setDefaultJoyAxisMapping(Event::JoystickOneDown, mode, 1, 1, 1);
+#define SET_DEFAULT_BTN(sdb_event, sdb_mode, sdb_stick, sdb_button, sdb_cmp_event) \
+  if(eraseAll || sdb_cmp_event == sdb_event) \
+    myEventHandler->addJoyButtonMapping(sdb_event, sdb_mode, sdb_stick, sdb_button, false);
 
-  mode = kMenuMode;  // Default menu/UI events
-  myEventHandler->setDefaultJoyAxisMapping(Event::UILeft, mode, 0, 0, 0);
-  myEventHandler->setDefaultJoyAxisMapping(Event::UIRight, mode, 0, 0, 1);
-  myEventHandler->setDefaultJoyAxisMapping(Event::UIUp, mode, 0, 1, 0);
-  myEventHandler->setDefaultJoyAxisMapping(Event::UIDown, mode, 0, 1, 1);
-}
+#define SET_DEFAULT_HAT(sdh_event, sdh_mode, sdh_stick, sdh_hat, sdh_dir, sdh_cmp_event) \
+  if(eraseAll || sdh_cmp_event == sdh_event) \
+    myEventHandler->addJoyHatMapping(sdh_event, sdh_mode, sdh_stick, sdh_hat, sdh_dir, false);
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void OSystem::setDefaultJoyHatMap()
-{
-// FIXME - add emul and UI events
+  bool eraseAll = (event == Event::NoType);
+  switch(mode)
+  {
+    case kEmulationMode:  // Default emulation events
+      // Left joystick left/right directions (assume joystick zero)
+      SET_DEFAULT_AXIS(Event::JoystickZeroLeft, mode, 0, 0, 0, event);
+      SET_DEFAULT_AXIS(Event::JoystickZeroRight, mode, 0, 0, 1, event);
+      // Left joystick up/down directions (assume joystick zero)
+      SET_DEFAULT_AXIS(Event::JoystickZeroUp, mode, 0, 1, 0, event);
+      SET_DEFAULT_AXIS(Event::JoystickZeroDown, mode, 0, 1, 1, event);
+      // Right joystick left/right directions (assume joystick one)
+      SET_DEFAULT_AXIS(Event::JoystickOneLeft, mode, 1, 0, 0, event);
+      SET_DEFAULT_AXIS(Event::JoystickOneRight, mode, 1, 0, 1, event);
+      // Right joystick left/right directions (assume joystick one)
+      SET_DEFAULT_AXIS(Event::JoystickOneUp, mode, 1, 1, 0, event);
+      SET_DEFAULT_AXIS(Event::JoystickOneDown, mode, 1, 1, 1, event);
+
+      // Left joystick (assume joystick zero, button zero)
+      SET_DEFAULT_BTN(Event::JoystickZeroFire1, mode, 0, 0, event);
+      // Right joystick (assume joystick one, button zero)
+      SET_DEFAULT_BTN(Event::JoystickOneFire1, mode, 1, 0, event);
+
+      // Left joystick left/right directions (assume joystick zero and hat 0)
+      SET_DEFAULT_HAT(Event::JoystickZeroLeft, mode, 0, 0, EVENT_HATLEFT, event);
+      SET_DEFAULT_HAT(Event::JoystickZeroRight, mode, 0, 0, EVENT_HATRIGHT, event);
+      // Left joystick up/down directions (assume joystick zero and hat 0)
+      SET_DEFAULT_HAT(Event::JoystickZeroUp, mode, 0, 0, EVENT_HATUP, event);
+      SET_DEFAULT_HAT(Event::JoystickZeroDown, mode, 0, 0, EVENT_HATDOWN, event);
+
+      break;
+
+    case kMenuMode:  // Default menu/UI events
+      SET_DEFAULT_AXIS(Event::UILeft, mode, 0, 0, 0, event);
+      SET_DEFAULT_AXIS(Event::UIRight, mode, 0, 0, 1, event);
+      SET_DEFAULT_AXIS(Event::UIUp, mode, 0, 1, 0, event);
+      SET_DEFAULT_AXIS(Event::UIDown, mode, 0, 1, 1, event);
+
+      // Left joystick (assume joystick zero, button zero)
+      SET_DEFAULT_BTN(Event::UISelect, mode, 0, 0, event);
+      // Right joystick (assume joystick one, button zero)
+      SET_DEFAULT_BTN(Event::UISelect, mode, 1, 0, event);
+
+      SET_DEFAULT_HAT(Event::UILeft, mode, 0, 0, EVENT_HATLEFT, event);
+      SET_DEFAULT_HAT(Event::UIRight, mode, 0, 0, EVENT_HATRIGHT, event);
+      SET_DEFAULT_HAT(Event::UIUp, mode, 0, 0, EVENT_HATUP, event);
+      SET_DEFAULT_HAT(Event::UIDown, mode, 0, 0, EVENT_HATDOWN, event);
+
+      break;
+
+    default:
+      break;
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -854,14 +960,6 @@ void OSystem::pollEvent()
 {
 }
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool OSystem::joyButtonHandled(int button)
-{
-  // Since we don't do any platform-specific event polling,
-  // no button is ever handled at this level
-  return false;
-}
-
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void OSystem::stateChanged(EventHandler::State state)
 {
@@ -871,12 +969,17 @@ void OSystem::stateChanged(EventHandler::State state)
 uInt64 OSystem::getTicks() const
 {
 #ifdef HAVE_GETTIMEOFDAY
+  // Gettimeofday natively refers to the UNIX epoch (a set time in the past)
   timeval now;
   gettimeofday(&now, 0);
 
   return uInt64(now.tv_sec) * 1000000 + now.tv_usec;
 #else
-  return uInt64(SDL_GetTicks()) * 1000;
+  // We use SDL_GetTicks, but add in the time when the application was
+  // initialized.  This is necessary, since SDL_GetTicks only measures how
+  // long SDL has been running, which can be the same between multiple runs
+  // of the application.
+  return uInt64(SDL_GetTicks() + myMillisAtStart) * 1000;
 #endif
 }
 
@@ -895,6 +998,15 @@ void OSystem::mainLoop()
       myTimingInfo.current = getTicks();
       myTimingInfo.virt += myTimePerFrame;
 
+      // Timestamps may periodically go out of sync, particularly on systems
+      // that can have 'negative time' (ie, when the time seems to go backwards)
+      // This normally results in having a very large delay time, so we check
+      // for that and reset the timers when appropriate
+      if((myTimingInfo.virt - myTimingInfo.current) > (myTimePerFrame << 1))
+      {
+        myTimingInfo.start = myTimingInfo.current = myTimingInfo.virt = getTicks();
+      }
+
       if(myTimingInfo.current < myTimingInfo.virt)
         SDL_Delay((myTimingInfo.virt - myTimingInfo.current) / 1000);
 
@@ -931,9 +1043,22 @@ bool OSystem::queryVideoHardware()
       return false;
 
   // First get the maximum windowed desktop resolution
-  const SDL_VideoInfo* info = SDL_GetVideoInfo();
-  myDesktopWidth  = info->current_w;
-  myDesktopHeight = info->current_h;
+  // Check the 'maxres' setting, which is an undocumented developer feature
+  // that specifies the desktop size
+  // Normally, this wouldn't be set, and we ask SDL directly
+  int w, h;
+  mySettings->getSize("maxres", w, h);
+  if(w <= 0 || h <= 0)
+  {
+    const SDL_VideoInfo* info = SDL_GetVideoInfo();
+    myDesktopWidth  = info->current_w;
+    myDesktopHeight = info->current_h;
+  }
+  else
+  {
+    myDesktopWidth  = BSPF_max(w, 320);
+    myDesktopHeight = BSPF_max(h, 240);
+  }
 
   // Various parts of the codebase assume a minimum screen size of 320x240
   assert(myDesktopWidth >= 320 && myDesktopHeight >= 240);
@@ -953,15 +1078,35 @@ bool OSystem::queryVideoHardware()
   }
   else
   {
+    // All modes must fit between the lower and upper limits of the desktop
+    // For 'small' desktop, this means larger than 320x240
+    // For 'large'/normal desktop, exclude all those less than 640x480
+    bool largeDesktop = myDesktopWidth >= 640 && myDesktopHeight >= 480;
+    uInt32 lowerWidth  = largeDesktop ? 640 : 320,
+           lowerHeight = largeDesktop ? 480 : 240;
     for(uInt32 i = 0; modes[i]; ++i)
+    {
+      if(modes[i]->w >= lowerWidth && modes[i]->w <= myDesktopWidth &&
+         modes[i]->h >= lowerHeight && modes[i]->h <= myDesktopHeight)
+      {
+        Resolution r;
+        r.width  = modes[i]->w;
+        r.height = modes[i]->h;
+        buf.str("");
+        buf << r.width << "x" << r.height;
+        r.name = buf.str();
+        myResolutions.insert_at(0, r);  // insert in opposite (of descending) order
+      }
+    }
+    // If no modes were valid, use the desktop dimensions
+    if(myResolutions.size() == 0)
     {
       Resolution r;
-      r.width  = modes[i]->w;
-      r.height = modes[i]->h;
-      buf.str("");
+      r.width  = myDesktopWidth;
+      r.height = myDesktopHeight;
       buf << r.width << "x" << r.height;
       r.name = buf.str();
-      myResolutions.insert_at(0, r);  // insert in opposite (of descending) order
+      myResolutions.push_back(r);
     }
   }
 
@@ -972,12 +1117,12 @@ bool OSystem::queryVideoHardware()
 /*
   Palette is defined as follows:
     // Base colors
-    kColor         TODO
-    kBGColor       TODO
+    kColor            Normal foreground color (non-text)
+    kBGColor          Normal background color (non-text)
     kShadowColor      Item is disabled
     kTextColor        Normal text color
     kTextColorHi      Highlighted text color
-    kTextColorEm   TODO
+    kTextColorEm      Emphasized text color
 
     // UI elements (dialog and widgets)
     kDlgColor         Dialog background
