@@ -8,12 +8,13 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2009 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2011 by Bradford W. Mott, Stephen Anthony
+// and the Stella Team
 //
-// See the file "license" for information on usage and redistribution of
+// See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: Settings.cxx 1852 2009-08-07 11:45:14Z stephena $
+// $Id: Settings.cxx 2301 2011-12-28 00:25:19Z stephena $
 //============================================================================
 
 #include <cassert>
@@ -40,30 +41,23 @@ Settings::Settings(OSystem* osystem)
 
   // OpenGL specific options
   setInternal("gl_filter", "nearest");
-  setInternal("gl_aspectn", "100");
+  setInternal("gl_aspectn", "90");
   setInternal("gl_aspectp", "100");
-  setInternal("gl_fsmax", "false");
+  setInternal("gl_fsmax", "true");
   setInternal("gl_lib", "libGL.so");
-  setInternal("gl_vsync", "false");
-  setInternal("gl_texrect", "false");
-//  setInternal("gl_accel", "true");
+  setInternal("gl_vsync", "true");
+  setInternal("gl_vbo", "true");
 
   // Framebuffer-related options
   setInternal("tia_filter", "zoom2x");
   setInternal("fullscreen", "0");
   setInternal("fullres", "auto");
-  setInternal("center", "true");
-  setInternal("grabmouse", "false");
+  setInternal("center", "false");
+  setInternal("grabmouse", "true");
   setInternal("palette", "standard");
-  setInternal("colorloss", "false");
+  setInternal("colorloss", "true");
   setInternal("timing", "sleep");
-
-  // TV filter options
-  setInternal("tv_tex", "off");
-  setInternal("tv_bleed", "off");
-  setInternal("tv_noise", "off");
-//  setInternal("tv_curve", "false");  // not yet implemented
-  setInternal("tv_phos", "false");
+  setInternal("uimessages", "true");
 
   // Sound options
   setInternal("sound", "true");
@@ -76,32 +70,36 @@ Settings::Settings(OSystem* osystem)
   // Input event options
   setInternal("keymap", "");
   setInternal("joymap", "");
-  setInternal("joyaxismap", "");
-  setInternal("joyhatmap", "");
-  setInternal("joydeadzone", "0");
+  setInternal("combomap", "");
+  setInternal("joydeadzone", "13");
   setInternal("joyallow4", "false");
-  setInternal("pspeed", "6");
+  setInternal("mcontrol", "auto");
+  setInternal("dsense", "5");
+  setInternal("msense", "7");
   setInternal("sa1", "left");
   setInternal("sa2", "right");
+  setInternal("ctrlcombo", "true");
 
   // Snapshot options
-  setInternal("ssdir", "");
+  setInternal("snapdir", "");
   setInternal("sssingle", "false");
   setInternal("ss1x", "false");
+  setInternal("ssinterval", "2");
 
   // Config files and paths
-  setInternal("romdir", "~");
+  setInternal("romdir", "");
   setInternal("statedir", "");
   setInternal("cheatfile", "");
   setInternal("palettefile", "");
   setInternal("propsfile", "");
   setInternal("eepromdir", "");
+  setInternal("cfgdir", "");
 
   // ROM browser options
   setInternal("uselauncher", "true");
   setInternal("launcherres", "640x480");
   setInternal("launcherfont", "medium");
-  setInternal("launcherexts", "allfiles");
+  setInternal("launcherexts", "allroms");
   setInternal("romviewer", "0");
   setInternal("lastrom", "");
 
@@ -113,13 +111,23 @@ Settings::Settings(OSystem* osystem)
 
   // Misc options
   setInternal("autoslot", "false");
-  setInternal("showinfo", "false");
-  setInternal("tiafloat", "true");
+  setInternal("loglevel", "1");
+  setInternal("logtoconsole", "0");
+  setInternal("tiadriven", "false");
+  setInternal("ramrandom", "true");
   setInternal("avoxport", "");
   setInternal("stats", "false");
-  setInternal("audiofirst", "true");
   setInternal("fastscbios", "false");
   setExternal("romloadcount", "0");
+  setExternal("maxres", "");
+
+  // Debugger options
+  setInternal("resolvedata", "auto");
+  setInternal("gfxformat", "2");
+  setInternal("showaddr", "true");
+
+  // Thumb ARM emulation options
+  setInternal("thumb.trapfatal", "true");
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -138,7 +146,7 @@ void Settings::loadConfig()
   ifstream in(myOSystem->configFile().c_str());
   if(!in || !in.is_open())
   {
-    cout << "ERROR: Couldn't load settings file\n";
+    myOSystem->logMessage("ERROR: Couldn't load settings file\n", 0);
     return;
   }
 
@@ -201,19 +209,32 @@ string Settings::loadCommandLine(int argc, char** argv)
         continue;
       }
 
+      ostringstream buf;
       if(++i >= argc)
       {
-        cerr << "Missing argument for '" << key << "'" << endl;
+        buf << "Missing argument for '" << key << "'" << endl;
+        myOSystem->logMessage(buf.str(), 0);
         return "";
       }
       string value = argv[i];
 
+      buf.str("");
+      buf << "  key = '" << key << "', value = '" << value << "'";
+
       // Settings read from the commandline must not be saved to 
       // the rc-file, unless they were previously set
       if(int idx = getInternalPos(key) != -1)
+      {
         setInternal(key, value, idx);   // don't set initialValue here
+        buf << "(I)\n";
+      }
       else
+      {
         setExternal(key, value);
+        buf << "(E)\n";
+      }
+
+      myOSystem->logMessage(buf.str(), 2);
     }
     else
       return key;
@@ -257,9 +278,28 @@ void Settings::validate()
   if(i < 0)        setInternal("joydeadzone", "0");
   else if(i > 29)  setInternal("joydeadzone", "29");
 
-  i = getInt("pspeed");
-  if(i < 1)        setInternal("pspeed", "1");
-  else if(i > 15)  setInternal("pspeed", "15");
+  s = getString("mcontrol");
+  if(s != "auto")
+  {
+    // Note: these constants are from Controller::MouseAxisType enum
+    if(s.length() != 2 || s[0] < '0' || s[0] > '5' || s[1] < '0' || s[1] > '5')
+      setInternal("mcontrol", "auto");
+  }
+
+  if(i < 1)        setInternal("dsense", "1");
+  else if(i > 10)  setInternal("dsense", "10");
+
+  i = getInt("dsense");
+  if(i < 1)        setInternal("dsense", "1");
+  else if(i > 10)  setInternal("dsense", "10");
+
+  i = getInt("msense");
+  if(i < 1)        setInternal("msense", "1");
+  else if(i > 15)  setInternal("msense", "15");
+
+  i = getInt("ssinterval");
+  if(i < 1)        setInternal("ssinterval", "2");
+  else if(i > 10)  setInternal("ssinterval", "10");
 
   s = getString("palette");
   if(s != "standard" && s != "z26" && s != "user")
@@ -273,23 +313,14 @@ void Settings::validate()
   if(i < 0)       setInternal("romviewer", "0");
   else if(i > 2)  setInternal("romviewer", "2");
 
-  s = getString("tv_tex");
-  if(s != "normal" && s != "stag")
-    setInternal("tv_tex", "off");
-
-  s = getString("tv_bleed");
-  if(s != "low" && s != "medium" && s != "high")
-    setInternal("tv_bleed", "off");
-
-  s = getString("tv_noise");
-  if(s != "low" && s != "medium" && s != "high")
-    setInternal("tv_noise", "off");
+  i = getInt("loglevel");
+  if(i < 0 || i > 2)
+    setInternal("loglevel", "1");
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void Settings::usage()
 {
-#ifndef MAC_OSX
   cout << endl
     << "Stella version " << STELLA_VERSION << endl
     << endl
@@ -311,21 +342,8 @@ void Settings::usage()
     << "  -gl_aspectn   <number>       Scale the TIA width by the given percentage in NTSC mode\n"
     << "  -gl_aspectp   <number>       Scale the TIA width by the given percentage in PAL mode\n"
     << "  -gl_fsmax     <1|0>          Stretch GL image in fullscreen emulation mode\n"
-    << "  -gl_vsync     <1|0>          Enable synchronize to vertical blank interrupt\n"
-    << "  -gl_texrect   <1|0>          Enable GL_TEXTURE_RECTANGLE extension\n"
-//    << "  -gl_accel     <1|0>          Enable SDL_GL_ACCELERATED_VISUAL\n"
-    << "  -tv_tex       <off|type>     OpenGL TV texturing, type is one of the following:\n"
-    << "                 normal         Aligned in a grid\n"
-    << "                 stag           Aligned in a staggered grid\n"
-    << "  -tv_bleed     <off|type>     OpenGL TV color bleed, type is one of the following:\n"
-    << "                 low            \n"
-    << "                 medium         \n"
-    << "                 high           \n"
-    << "  -tv_noise     <off|type>     OpenGL TV RF noise emulation, type is one of the following:\n"
-    << "                 low            \n"
-    << "                 medium         \n"
-    << "                 high           \n"
-    << "  -tv_phos      <1|0>          OpenGL TV phosphor burn-off\n"
+    << "  -gl_vsync     <1|0>          Enable 'synchronize to vertical blank interrupt'\n"
+    << "  -gl_vbo       <1|0>          Enable 'vertex buffer objects'\n"
     << endl
   #endif
     << "  -tia_filter   <filter>       Use the specified filter in emulation mode\n"
@@ -339,6 +357,7 @@ void Settings::usage()
     << "  -colorloss    <1|0>          Enable PAL color-loss effect\n"
     << "  -framerate    <number>       Display the given number of frames per second (0 to auto-calculate)\n"
     << "  -timing       <sleep|busy>   Use the given type of wait between frames\n"
+    << "  -uimessages   <1|0>          Show onscreen UI messages for different events\n"
     << endl
   #ifdef SOUND_SUPPORT
     << "  -sound        <1|0>          Enable sound generation\n"
@@ -350,18 +369,23 @@ void Settings::usage()
     << endl
   #endif
     << "  -cheat        <code>         Use the specified cheatcode (see manual for description)\n"
-    << "  -showinfo     <1|0>          Shows some game info on commandline\n"
+    << "  -loglevel     <0|1|2>        Set level of logging during application run\n"
+    << "  -logtoconsole <1|0>          Log output to console/commandline\n"
     << "  -joydeadzone  <number>       Sets 'deadzone' area for analog joysticks (0-29)\n"
     << "  -joyallow4    <1|0>          Allow all 4 directions on a joystick to be pressed simultaneously\n"
-    << "  -pspeed       <number>       Speed of digital emulated paddle movement (1-15)\n"
+    << "  -mcontrol     <auto|XY>      Use mouse axes as specified paddle controller type (see manual)\n"
+    << "  -dsense       <number>       Sensitivity of digital emulated paddle movement (1-10)\n"
+    << "  -msense       <number>       Sensitivity of mouse emulated paddle movement (1-15)\n"
     << "  -sa1          <left|right>   Stelladaptor 1 emulates specified joystick port\n"
     << "  -sa2          <left|right>   Stelladaptor 2 emulates specified joystick port\n"
+    << "  -ctrlcombo    <1|0>          Use key combos involving the Control key (Control-Q for quit may be disabled!)\n"
     << "  -autoslot     <1|0>          Automatically switch to next save slot when state saving\n"
-    << "  -audiofirst   <1|0>          Initial audio before video (required for some ATI video cards)\n"
+    << "  -stats        <1|0>          Overlay console info during emulation\n"
     << "  -fastscbios   <1|0>          Disable Supercharger BIOS progress loading bars\n"
-    << "  -ssdir        <path>         The directory to save snapshot files to\n"
+    << "  -snapdir      <path>         The directory to save snapshot files to\n"
     << "  -sssingle     <1|0>          Generate single snapshot instead of many\n"
     << "  -ss1x         <1|0>          Generate TIA snapshot in 1x mode (ignore scaling/effects)\n"
+    << "  -ssinterval   <number        Number of seconds between snapshots in continuous snapshot mode\n"
     << endl
     << "  -rominfo      <rom>          Display detailed information for the given ROM\n"
     << "  -listrominfo                 Display contents of stella.pro, one line per ROM entry\n"
@@ -381,21 +405,27 @@ void Settings::usage()
     << "  -palettefile  <file>         Full pathname of user-defined palette file\n"
     << "  -propsfile    <file>         Full pathname of ROM properties file\n"
     << "  -eepromdir    <dir>          Directory in which to save EEPROM files\n"
+    << "  -cfgdir       <dir>          Directory in which to save Distella config files\n"
     << "  -avoxport     <name>         The name of the serial port where an AtariVox is connected\n"
+    << "  -maxres       <WxH>          Used by developers to force the maximum size of the application window\n"
+    << "  -holdreset                   Start the emulator with the Game Reset switch held down\n"
+    << "  -holdselect                  Start the emulator with the Game Select switch held down\n"
+    << "  -holdbutton0                 Start the emulator with the left joystick button held down\n"
+    << "  -tiadriven    <1|0>          Drive unused TIA pins randomly on a read/peek\n"
+    << "  -ramrandom    <1|0>          Randomize the contents of RAM on reset\n"
     << "  -help                        Show the text you're now reading\n"
   #ifdef DEBUGGER_SUPPORT
     << endl
     << " The following options are meant for developers\n"
     << " Arguments are more fully explained in the manual\n"
     << endl
+    << "   -resolvedata  <never|       Set automatic code vs. data determination in the disassembler\n"
+    << "                  always|auto>\n"
+    << "   -gfxformat    <2|16>        Set base to use for displaying GFX sections in the disassembler\n"
+    << "   -showaddr     <1|0>         Show opcode addresses in the disassembler\n"
     << "   -debuggerres  <WxH>         The resolution to use in debugger mode\n"
     << "   -break        <address>     Set a breakpoint at 'address'\n"
     << "   -debug                      Start in debugger mode\n"
-    << "   -holdreset                  Start the emulator with the Game Reset switch held down\n"
-    << "   -holdselect                 Start the emulator with the Game Select switch held down\n"
-    << "   -holdbutton0                Start the emulator with the left joystick button held down\n"
-    << "   -stats        <1|0>         Overlay console info during emulation\n"
-    << "   -tiafloat     <1|0>         Set unused TIA pins floating on a read/peek\n"
     << endl
     << "   -bs          <arg>          Sets the 'Cartridge.Type' (bankswitch) property\n"
     << "   -type        <arg>          Same as using -bs\n"
@@ -414,8 +444,7 @@ void Settings::usage()
     << "   -pp          <arg>          Sets the 'Display.Phosphor' property\n"
     << "   -ppblend     <arg>          Sets the 'Display.PPBlend' property\n"
   #endif
-    << endl;
-#endif
+    << endl << flush;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -439,7 +468,7 @@ void Settings::saveConfig()
   ofstream out(myOSystem->configFile().c_str());
   if(!out || !out.is_open())
   {
-    cout << "ERROR: Couldn't save settings file\n";
+    myOSystem->logMessage("ERROR: Couldn't save settings file\n", 0);
     return;
   }
 
@@ -516,7 +545,8 @@ void Settings::setString(const string& key, const string& value)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void Settings::getSize(const string& key, int& x, int& y) const
 {
-  char c;
+  char c = '\0';
+  x = y = -1;
   string size = getString(key);
   istringstream buf(size);
   buf >> x >> c >> y;
