@@ -8,12 +8,13 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2009 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2011 by Bradford W. Mott, Stephen Anthony
+// and the Stella Team
 //
-// See the file "license" for information on usage and redistribution of
+// See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: FileSnapDialog.cxx 1779 2009-06-18 17:38:42Z stephena $
+// $Id: FileSnapDialog.cxx 2292 2011-12-22 21:29:24Z stephena $
 //
 //   Based on code from ScummVM - Scumm Interpreter
 //   Copyright (C) 2002-2004 The ScummVM project
@@ -25,6 +26,7 @@
 #include "EditTextWidget.hxx"
 #include "FSNode.hxx"
 #include "LauncherDialog.hxx"
+#include "PopUpWidget.hxx"
 #include "Settings.hxx"
 
 #include "FileSnapDialog.hxx"
@@ -32,7 +34,8 @@
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 FileSnapDialog::FileSnapDialog(
       OSystem* osystem, DialogContainer* parent,
-      const GUI::Font& font, GuiObject* boss)
+      const GUI::Font& font, GuiObject* boss,
+      int max_w, int max_h)
   : Dialog(osystem, parent, 0, 0, 0, 0),
     CommandSender(boss),
     myBrowser(NULL),
@@ -49,7 +52,7 @@ FileSnapDialog::FileSnapDialog(
 
   // Set real dimensions
   _w = 52 * fontWidth + 8;
-  _h = 11 * (lineHeight + 4) + 10;
+  _h = 12 * (lineHeight + 4) + 10;
 
   xpos = vBorder;  ypos = vBorder;
 
@@ -63,15 +66,15 @@ FileSnapDialog::FileSnapDialog(
                                  _w - xpos - 10, lineHeight, "");
   wid.push_back(myRomPath);
 
-  // State directory
+  // Snapshot path
   xpos = vBorder;  ypos += romButton->getHeight() + 3;
   b = new ButtonWidget(this, font, xpos, ypos, buttonWidth, buttonHeight,
-                       "State path:", kChooseStateDirCmd);
+                       "Snapshot path:", kChooseSnapDirCmd);
   wid.push_back(b);
   xpos += buttonWidth + 10;
-  myStatePath = new EditTextWidget(this, font, xpos, ypos + 2,
-                                   _w - xpos - 10, lineHeight, "");
-  wid.push_back(myStatePath);
+  mySnapPath = new EditTextWidget(this, font, xpos, ypos + 2,
+                                  _w - xpos - 10, lineHeight, "");
+  wid.push_back(mySnapPath);
 
   // Cheat file
   xpos = vBorder;  ypos += b->getHeight() + 3;
@@ -103,15 +106,15 @@ FileSnapDialog::FileSnapDialog(
                                    _w - xpos - 10, lineHeight, "");
   wid.push_back(myPropsFile);
 
-  // Snapshot path
+  // State directory
   xpos = vBorder;  ypos += b->getHeight() + 3;
   b = new ButtonWidget(this, font, xpos, ypos, buttonWidth, buttonHeight,
-                       "Snapshot path:", kChooseSnapDirCmd);
+                       "State path:", kChooseStateDirCmd);
   wid.push_back(b);
   xpos += buttonWidth + 10;
-  mySnapPath = new EditTextWidget(this, font, xpos, ypos + 2,
-                                  _w - xpos - 10, lineHeight, "");
-  wid.push_back(mySnapPath);
+  myStatePath = new EditTextWidget(this, font, xpos, ypos + 2,
+                                   _w - xpos - 10, lineHeight, "");
+  wid.push_back(myStatePath);
 
   // EEPROM directory
   xpos = vBorder;  ypos += b->getHeight() + 3;
@@ -126,7 +129,7 @@ FileSnapDialog::FileSnapDialog(
   // Snapshot single or multiple saves
   xpos = 30;  ypos += b->getHeight() + 5;
   mySnapSingle = new CheckboxWidget(this, font, xpos, ypos,
-                                    "Multiple snapshots");
+                                    "Overwrite snapshots");
   wid.push_back(mySnapSingle);
 
   // Snapshot in 1x mode (ignore scaling)
@@ -135,6 +138,25 @@ FileSnapDialog::FileSnapDialog(
                                 "Snapshot in 1x mode");
   wid.push_back(mySnap1x);
 
+  // Snapshot interval (continuous mode)
+  StringMap items;
+  items.clear();
+  items.push_back("1 second", "1");
+  items.push_back("2 seconds", "2");
+  items.push_back("3 seconds", "3");
+  items.push_back("4 seconds", "4");
+  items.push_back("5 seconds", "5");
+  items.push_back("6 seconds", "6");
+  items.push_back("7 seconds", "7");
+  items.push_back("8 seconds", "8");
+  items.push_back("9 seconds", "9");
+  items.push_back("10 seconds", "10");
+  xpos = 30;  ypos += b->getHeight();
+  mySnapInterval = new PopUpWidget(this, font, xpos, ypos,
+                                   font.getStringWidth("10 seconds"), lineHeight,
+                                   items, "Continuous snapshot interval: ");
+  wid.push_back(mySnapInterval);
+
   // Add Defaults, OK and Cancel buttons
   b = new ButtonWidget(this, font, 10, _h - buttonHeight - 10,
                        font.getStringWidth("Defaults") + 20, buttonHeight,
@@ -152,7 +174,7 @@ FileSnapDialog::FileSnapDialog(
   }
 
   // Create file browser dialog
-  myBrowser = new BrowserDialog(this, font);
+  myBrowser = new BrowserDialog(this, font, max_w, max_h);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -166,28 +188,30 @@ void FileSnapDialog::loadConfig()
 {
   const Settings& settings = instance().settings();
   myRomPath->setEditString(settings.getString("romdir"));
-  myStatePath->setEditString(settings.getString("statedir"));
+  mySnapPath->setEditString(settings.getString("snapdir"));
   myCheatFile->setEditString(settings.getString("cheatfile"));
   myPaletteFile->setEditString(settings.getString("palettefile"));
   myPropsFile->setEditString(settings.getString("propsfile"));
-  mySnapPath->setEditString(settings.getString("ssdir"));
   myEEPROMPath->setEditString(settings.getString("eepromdir"));
-  mySnapSingle->setState(!settings.getBool("sssingle"));
+  myStatePath->setEditString(settings.getString("statedir"));
+  mySnapSingle->setState(settings.getBool("sssingle"));
   mySnap1x->setState(settings.getBool("ss1x"));
+  mySnapInterval->setSelected(instance().settings().getString("ssinterval"), "2");
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FileSnapDialog::saveConfig()
 {
   instance().settings().setString("romdir", myRomPath->getEditString());
-  instance().settings().setString("statedir", myStatePath->getEditString());
+  instance().settings().setString("snapdir", mySnapPath->getEditString());
   instance().settings().setString("cheatfile", myCheatFile->getEditString());
   instance().settings().setString("palettefile", myPaletteFile->getEditString());
   instance().settings().setString("propsfile", myPropsFile->getEditString());
-  instance().settings().setString("ssdir", mySnapPath->getEditString());
+  instance().settings().setString("statedir", myStatePath->getEditString());
   instance().settings().setString("eepromdir", myEEPROMPath->getEditString());
-  instance().settings().setBool("sssingle", !mySnapSingle->getState());
+  instance().settings().setBool("sssingle", mySnapSingle->getState());
   instance().settings().setBool("ss1x", mySnap1x->getState());
+  instance().settings().setString("ssinterval", mySnapInterval->getSelectedTag());
 
   // Flush changes to disk and inform the OSystem
   instance().settings().saveConfig();
@@ -198,17 +222,13 @@ void FileSnapDialog::saveConfig()
 void FileSnapDialog::setDefaults()
 {
   FilesystemNode node;
-  string basedir = instance().baseDir();
-  if(basedir.compare(basedir.length()-1, 1, BSPF_PATH_SEPARATOR, 0, 1) != 0)
-    basedir.append(BSPF_PATH_SEPARATOR);
+  const string& basedir = instance().baseDir();
 
   node = FilesystemNode("~");
   myRomPath->setEditString(node.getRelativePath());
 
-  const string& statedir = basedir + "state";
-  node = FilesystemNode(statedir);
-  myStatePath->setEditString(node.getRelativePath());
-
+  mySnapPath->setEditString(instance().defaultSnapDir());
+ 
   const string& cheatfile = basedir + "stella.cht";
   node = FilesystemNode(cheatfile);
   myCheatFile->setEditString(node.getRelativePath());
@@ -225,12 +245,13 @@ void FileSnapDialog::setDefaults()
   node = FilesystemNode(eepromdir);
   myEEPROMPath->setEditString(node.getRelativePath());
 
-  const string& ssdir = basedir + "snapshots";
-  node = FilesystemNode(ssdir);
-  mySnapPath->setEditString(node.getRelativePath());
+  const string& statedir = basedir + "state";
+  node = FilesystemNode(statedir);
+  myStatePath->setEditString(node.getRelativePath());
 
-  mySnapSingle->setState(true);
+  mySnapSingle->setState(false);
   mySnap1x->setState(false);
+  mySnapInterval->setSelected("2", "2");
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -255,9 +276,9 @@ void FileSnapDialog::handleCommand(CommandSender* sender, int cmd,
                       FilesystemNode::kListDirectoriesOnly, kRomDirChosenCmd);
       break;
 
-    case kChooseStateDirCmd:
-      myBrowser->show("Select state directory:", myStatePath->getEditString(),
-                      FilesystemNode::kListDirectoriesOnly, kStateDirChosenCmd);
+    case kChooseSnapDirCmd:
+      myBrowser->show("Select snapshot directory:", mySnapPath->getEditString(),
+                      FilesystemNode::kListDirectoriesOnly, kSnapDirChosenCmd);
       break;
 
     case kChooseCheatFileCmd:
@@ -275,16 +296,16 @@ void FileSnapDialog::handleCommand(CommandSender* sender, int cmd,
                       FilesystemNode::kListAll, kPropsFileChosenCmd);
       break;
 
-    case kChooseSnapDirCmd:
-      myBrowser->show("Select snapshot directory:", mySnapPath->getEditString(),
-                      FilesystemNode::kListDirectoriesOnly, kSnapDirChosenCmd);
-      break;
-
     case kChooseEEPROMDirCmd:
       myBrowser->show("Select EEPROM directory:", myEEPROMPath->getEditString(),
                       FilesystemNode::kListDirectoriesOnly, kEEPROMDirChosenCmd);
       break;
 
+    case kChooseStateDirCmd:
+      myBrowser->show("Select state directory:", myStatePath->getEditString(),
+                      FilesystemNode::kListDirectoriesOnly, kStateDirChosenCmd);
+      break;
+
     case kRomDirChosenCmd:
     {
       FilesystemNode dir(myBrowser->getResult());
@@ -292,10 +313,10 @@ void FileSnapDialog::handleCommand(CommandSender* sender, int cmd,
       break;
     }
 
-    case kStateDirChosenCmd:
+    case kSnapDirChosenCmd:
     {
       FilesystemNode dir(myBrowser->getResult());
-      myStatePath->setEditString(dir.getRelativePath());
+      mySnapPath->setEditString(dir.getRelativePath());
       break;
     }
 
@@ -320,13 +341,6 @@ void FileSnapDialog::handleCommand(CommandSender* sender, int cmd,
       break;
     }
 
-    case kSnapDirChosenCmd:
-    {
-      FilesystemNode dir(myBrowser->getResult());
-      mySnapPath->setEditString(dir.getRelativePath());
-      break;
-    }
-
     case kEEPROMDirChosenCmd:
     {
       FilesystemNode dir(myBrowser->getResult());
@@ -334,6 +348,13 @@ void FileSnapDialog::handleCommand(CommandSender* sender, int cmd,
       break;
     }
 
+    case kStateDirChosenCmd:
+    {
+      FilesystemNode dir(myBrowser->getResult());
+      myStatePath->setEditString(dir.getRelativePath());
+      break;
+    }
+
     case kReloadRomDirCmd:
       sendCommand(kReloadRomDirCmd, 0, 0);	 
       break;
